/*------------------------------------------------------*
 | Author: Maurizio Loreti, aka MLO or (HAM) I3NOO      |
 | Work:   University of Padova - Department of Physics |
 |         Via F. Marzolo, 8 - 35131 PADOVA - Italy     |
 | Phone:  +39 (049) 827-7216   FAX: +39 (049) 827-7102 |
 | EMail:  loreti@padova.infn.it                        |
 | WWW:    http://wwwcdf.pd.infn.it/~loreti/mlo.html    |
 *------------------------------------------------------*

 File: config.c ; an interface to the use of configuration
       files.

 For every variable whose value has to be configured using
 a configuration file, a call is made to addConfig(); this
 procedure builds a double-linked list containing symbolic
 names and variable addresses.  Then readConfig() reads a
 configuration file, that may contain:
 - comments, started by ';' (to insert a literal semicolon
   in a string, use \;)
 - blank lines, without effect;
 - assignments, of the type name=value; assignments may be
   of numeric values or of strings and, for numeric values,
   the target may be an array.  Whitespace may be present
   before or after both keyword and value.

 Integer values may be expressed in octal, hexadecimal or
 decimal notation; string assignments assign from the first
 to the last non-whitespace character after '='.  When a
 value has been assigned to a variable, the corresponding
 node is removed from the linked list; so that more than
 one assignment to the same variable is not permitted.
 Numerical assignments to an array must be performed with
 successive values separated by ':' (with, as usual,
 optional whitespace).

 remConfig() checks for an empty list (and removes all the
 remaining nodes).

 $Id: config.c,v 1.8 2000/10/24 09:25:13 loreti Exp $

 *------------------------------------------------------*/

#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdarg.h>
#include <limits.h>

#include "config.h"

/**
 | Address of the first and the last node in the linked list
**/

static cNode *root = 0;
static cNode *last = 0;

/**
 | Local prototypes
**/

static int findConfig(const int, const char *, char *, const char *,
                      const unsigned int);
static char *getDelim(char **, char *);

/*--------------------------------------------------------------------*/

int addConfig(
  const int     print,          /* Printout wanted ? */
  cType         type,           /* Variable type,    */
  const char   *name,           /*  symbolic name,   */
  unsigned int  num,            /*   dimension,      */
  void         *target,         /*    and address.   */
  ...
) {
  cNode *pCN;

/**
 | Checks for an already known symbolic name the linked list, and for
 | an invalid type.
**/

  for (pCN = root;   pCN;   pCN = pCN->next) {
    if (!strcmp(pCN->name, name)) {
      if (print) {
        fprintf(stderr,
                "addConfig: variable \"%s\" already declared\n",
                name);
      }
      return CONFIG_ERROR;
    }
  }

  switch (type) {
    case CONFIG_STRING:
      if (num > 1) {
        type = CONFIG_STRARRAY;
      }
      break;

    case CONFIG_STRARRAY:
      if (num < 2) {
        type = CONFIG_STRING;
      }
      break;

    case CONFIG_LONG:
    case CONFIG_ULONG:
    case CONFIG_INT:
    case CONFIG_UINT:
    case CONFIG_DOUBLE:
      break;

    default:
      if (print) {
        fprintf(stderr,
                "addConfig: invalid type=%d for variable \"%s\"\n",
                type, name);
      }
      return CONFIG_ERROR;
  } /* switch */

/**
 | Allocates and initialises a new node, that is then inserted at
 | the _end_ of the linked list (FIFO).  The so-called "struct hack"
 | (not guaranteed by the ANSI/ISO standard) is used to insert the
 | symbolic name into the node.
**/

  if ((pCN = malloc(sizeof(cNode) + strlen(name))) == 0) {
    if (print) {
      fputs("addConfig: no heap memory available\n", stderr);
    }
    return CONFIG_ERROR;
  }

  if (num < 1) num = 1;         /* num=0 equivalent to num=1 */

  pCN->prev = last;
  pCN->next = 0;
  pCN->type = type;
  pCN->num  = num;
  strcpy(pCN->name, name);

/**
 | If the destination is an array of strings:
 | - allocates a temporary buffer where the right hand side of the
 |   assignment mey be stored, to be parsed afterwards (the address
 |   of this buffer is saved in pCN->target, to simulate a request
 |   for a STRING variable);
 | - allocates in pCN->auxTarget an array of "num" pointers to char,
 |   to save the addresses of the "num" target substrings.
**/

  if (type == CONFIG_STRARRAY) {
    int       i;
    va_list   vl;
    char     *pc;
    char    **to;

    pCN->target    = malloc(CONFIG_LINE_LEN);
    pCN->auxTarget = malloc(num * sizeof(char *));

    if (pCN->target == 0   ||   pCN->auxTarget == 0) {
      if (print) {
        fputs("addConfig: no heap memory available\n", stderr);
      }
      return CONFIG_ERROR;
    }

    va_start(vl, target);
    for (pc = target,  to = pCN->auxTarget,  i = 0;
         i < num;
         pc = va_arg(vl, char *), to++, i++) {
      *to = pc;
    }
    va_end(vl);
  } else {
    pCN->target    = target;
    pCN->auxTarget = 0;
  } /* if (type) */

/**
 | Inserts the node in the linked list
**/

  if (root) {
    last->next = pCN;
    last       = pCN;
  } else {
    root = last = pCN;
  }

  return CONFIG_OK;
}

int readConfig(
  const int   print,            /* Printout wanted ? */
  FILE       *pF,               /* File pointer and  */
  const char *fileName          /*  file name        */
) {
  char buffer[CONFIG_LINE_LEN];
  unsigned int lineNr = 1;
  int retVal          = CONFIG_OK;

  for (  ;  fgets(buffer, CONFIG_LINE_LEN, pF);  lineNr++) {
    char *pc;
    char *pName;
    char *pValue;

/**
 | For every line, strips the comments inserting a '\0' over the
 | first unquoted ';' found.
**/

    for (pc = buffer,  pName = buffer + 1;   *pc;   ) {
      if (*pc == CONFIG_QUOTE   &&  *pName == CONFIG_COMMENT) {
        strcpy(pc, pName);
      } else if (*pc == CONFIG_COMMENT) {
        *pc = '\0';
        break;
      }
      pc = pName++;
    }

/**
 | Skips the leading whitespace; "pName" points to the first non blank
 | character (or to the trailing '\0': an empty line).  The first non
 | blank character cannot be a '='.
**/

    for (pName=buffer; *pName && isspace((unsigned char) *pName); pName++) { }

    if (*pName == '\0') {
      continue;
    } else if (*pName == CONFIG_EQUAL) {
      if (print) {
        fprintf(stderr, "readConfig: file \"%s\", line %u: "
                "variable name missing in an assignment\n",
                fileName, lineNr);
      }
      retVal = CONFIG_ERROR;
      continue;
    }

/**
 | Looks for the assignment operator '='; delimits the symbolic name
 | field.
**/

    if ((pValue = strchr(pName, CONFIG_EQUAL)) == 0) {
      for (pc = pName;   *pc && !isspace((unsigned char) *pc);   pc++) { }
      *pc = '\0';
      if (print) {
        fprintf(stderr, "readConfig: file \"%s\", line %u: "
                "no assignment to variable \"%s\"\n",
                fileName, lineNr, pName);
      }
      retVal = CONFIG_ERROR;
      continue;
    }

    *pValue++ = '\0';
    for (pc = pName;   *pc && !isspace((unsigned char) *pc);   pc++) { }
    *pc = '\0';

/**
 | Skips the leading whitespace after the '=', to find the beginning
 | of the "value" field
**/

    while (*pValue && isspace((unsigned char) *pValue)) ++pValue;
    if (*pValue == 0) {
      if (print) {
        fprintf(stderr, "readConfig: file \"%s\", line %u: "
                "no value for variable \"%s\"\n",
                fileName, lineNr, pName);
      }
      retVal = CONFIG_ERROR;
      continue;
    }

/**
 | Calls findConfig() to perform the assignment
**/

    if (findConfig(print, pName, pValue, fileName, lineNr) != CONFIG_OK) {
      retVal = CONFIG_ERROR;
    }
  }

/**
 | Checks for read error or end-of-file
**/

  if (!feof(pF)) {
    if (print) {
      fprintf(stderr,
              "readConfig: error reading file \"%s\", line %u\n",
              fileName, lineNr);
      perror("readConfig");
    }
    return CONFIG_ERROR;
  } else {
    return retVal;
  }
}

int remConfig(
  const int print               /* Printout wanted ? */
) {
  cNode *pCN = root;

  while (pCN) {
    cNode *next = pCN->next;

    if (print) {
      fprintf(stderr,
              "remConfig: variable \"%s\" not initialised\n",
              pCN->name);
    }

    if (pCN->type == CONFIG_STRARRAY) {
      free(pCN->target);
      free(pCN->auxTarget);
    }
    free(pCN);
    pCN = next;
  }

  if (root) {
    root = last = 0;
    return CONFIG_ERROR;
  } else {
    return CONFIG_OK;
  }
}

static int findConfig(
  const int           print,    /* Printout wanted ? */
  const char         *pName,
        char         *pValue,
  const char         *fileName,
  const unsigned int  lineNr
) {
  cNode *pCN;

/**
 | Loops over the linked list, looking for the symbolic name found
**/

  for (pCN = root;   pCN;   pCN = pCN->next) {
    if (!strcmp(pName, pCN->name)) {

/**
 | Converts the string to a numeric value, and stores the result inside
 | the variable; for string assignments, copies to the destination
 | address up to the last non-whitespace character.
 |
 | If the target is an integer (signed or unsigned), first converts the
 | string to a long with strtol or strtoul (more reliable than atoi);
 | then checks that the result is compatible with an integer variable.
**/

      switch (pCN->type) {
        case CONFIG_LONG:
        {
          long int     *pl;
          char         *pc;
          unsigned int  num;

          pc  = getDelim(&pValue, CONFIG_SEP);
          num = pCN->num;
          pl  = (long *) pCN->target;

          while (num--) {
            long int l;
            char *pLast;

            if (pc == 0) goto warn;
            l = strtol(pc, &pLast, 0);
            if (!LAST_FIELD(pLast)) goto error;
            pc = getDelim(&pValue, CONFIG_SEP);
            *pl++ = l;
          }
        }
        break;

        case CONFIG_ULONG:
        {
          unsigned long int *pul;
          char              *pc;
          unsigned int       num;

          pc  = getDelim(&pValue, CONFIG_SEP);
          num = pCN->num;
          pul = (unsigned long *) pCN->target;

          while (num--) {
            unsigned long int ul;
            char *pLast;

            if (pc == 0) goto warn;
            ul = strtoul(pc, &pLast, 0);
            if (!LAST_FIELD(pLast)) goto error;
            pc = getDelim(&pValue, CONFIG_SEP);
            *pul++ = ul;
          }
        }
        break;

        case CONFIG_INT:
        {
          int          *pi;
          char         *pc;
          unsigned int  num;

          pc  = getDelim(&pValue, CONFIG_SEP);
          num = pCN->num;
          pi  = (int *) pCN->target;

          while (num--) {
            long int l;
            char *pLast;

            if (pc == 0) goto warn;
            l = strtol(pc, &pLast, 0);
            if (!LAST_FIELD(pLast)) goto error;
            if (l < INT_MIN   ||   l > INT_MAX) goto overflow;
            pc = getDelim(&pValue, CONFIG_SEP);
            *pi++ = (int) l;
          }
        }
        break;

        case CONFIG_UINT:
        {
          unsigned int *pui;
          char         *pc;
          unsigned int  num;

          pc  = getDelim(&pValue, CONFIG_SEP);
          num = pCN->num;
          pui = (unsigned int *) pCN->target;

          while (num--) {
            unsigned long int ul;
            char *pLast;

            if (pc == 0) goto warn;
            ul = strtoul(pc, &pLast, 0);
            if (!LAST_FIELD(pLast)) goto error;
            if (ul > UINT_MAX) goto overflow;
            pc = getDelim(&pValue, CONFIG_SEP);
            *pui++ = (unsigned int) ul;
          }
        }
        break;

        case CONFIG_DOUBLE:
        {
          double       *pd;
          char         *pc;
          unsigned int  num;

          pc  = getDelim(&pValue, CONFIG_SEP);
          num = pCN->num;
          pd  = (double *) pCN->target;

          while (num--) {
            double d;
            char *pLast;

            if (pc == 0) goto warn;
            d = strtod(pc, &pLast);
            if (!LAST_FIELD(pLast)) goto error;
            pc = getDelim(&pValue, CONFIG_SEP);
            *pd++ = d;
          }
        }
        break;

        case CONFIG_STRING:
        case CONFIG_STRARRAY:
        {
          char *pc;
          char *pLast = 0;

          for (pc = pValue;   *pc;   pc++) {
            if (!isspace((unsigned char) *pc)) pLast = pc;
          }
          for (pc = pCN->target;   pValue <= pLast;   ) {
            *pc++ = *pValue++;
          }
          *pc = '\0';
        }
        break;

        default:
          if (print) {
            fprintf(stderr, "Can't happen: file \"%s\", line %d\n",
                    __FILE__, __LINE__);
          }
          return CONFIG_ERROR;

/**
 | If too few values are given to initialise completely an array, the
 | uninitialised elements are left as they were; only a warning is
 | issued, and the assignment node destroyed.
**/

warn:
          if (print) {
            fprintf(stderr, "findConfig: file \"%s\", line %u: "
                    "too few values to initialise %s[%d]\n",
                    fileName, lineNr, pName, pCN->num);
          }
          break;
      } /* switch */

/**
 | If the destination is a string array, the whole right hand side of
 | the assignment has been stored into pCN->target; we must now parse
 | the substrings and store them in the appropriate character arrays.
**/

      if (pCN->type == CONFIG_STRARRAY) {
        char  *from;
        char **to;
        int    i;

        for (from = pCN->target,  to = pCN->auxTarget,  i = 0;
             i < pCN->num;
             to++,  i++) {
          char *pc;

          if ((pc = getDelim(&from, CONFIG_SEP)) == 0) break;
          strcpy(*to, pc);
        }

        if (i < pCN->num   &&   print) {
          fprintf(stderr, "findConfig: file \"%s\", line %u: "
                  "too few values to initialise %s[%d]\n",
                  fileName, lineNr, pName, pCN->num);
        }
      }

/**
 | In case of success, removes the current node from the linked list
**/

      if (pCN->prev) {
        pCN->prev->next = pCN->next;
      } else {
        root = pCN->next;
      }

      if (pCN->next) {
        pCN->next->prev = pCN->prev;
      } else {
        last = pCN->prev;
      }

      if (pCN->type == CONFIG_STRARRAY) {
        free(pCN->target);
        free(pCN->auxTarget);
      }
      free(pCN);
      return CONFIG_OK;
    } /* if (!strcmp) */
  } /* loop over nodes */

/**
 | Here only if the symbolic name does not match any of the nodes
**/

  if (print) {
    fprintf(stderr,
            "findConfig: file \"%s\", line %u: unknown variable \"%s\"\n",
            fileName, lineNr, pName);
  }
  return CONFIG_ERROR;

/**
 | Error exit for invalid input
**/

error:
  if (print) {
    fprintf(stderr, "findConfig: file \"%s\", line %u: "
            "invalid character in numeric field\n",
            fileName, lineNr);
  }
  return CONFIG_ERROR;

overflow:
  if (print) {
    fprintf(stderr, "findConfig: file \"%s\", line %u: "
            "value too large for an integer type\n",
            fileName, lineNr);
  }
  return CONFIG_ERROR;
}

char *getDelim(
  char **pStart,                /* Points to the start of the string */
  char  *sep                    /* Separator characters              */
){
  register char *retVal;
  register char *pc;
  register char *fb = 0;

/**
 | - Checks pStart, *pStart and sep: they cannot be NULL;
 | - skips leading whitespace;
 | - if the string's trailing zero has been reached, returns NULL
 |   (for 'no more tokens').
**/

  if (!pStart  ||  !(retVal = *pStart)  ||  !sep)       return 0;

  while (*retVal  &&  isspace((unsigned char) *retVal)) retVal++;

  if (!*(pc = retVal))                                  return 0;

/**
 | - Scans the token up to its end, the trailing zero or a separator:
 |   saves a pointer to the first whitespace found after a non-whitespace
 |   character (fb);
 | - if a separator has been found, it is substituted by a '\0';
 | - if whitespace has been found after the end of the token, a '\0' is
 |   stored there too.
**/

  while (*pc) {
    if (isspace((unsigned char) *pc)) {
      if (!fb) fb = pc;
    } else {
      if (strchr(sep, *pc)) {
        *pc++ = '\0';
        break;
      }
      fb = 0;
    }
    pc++;
  }

  if (fb) *fb = '\0';

  *pStart = pc;

  return retVal;
}
