//     Author: Maurizio Loreti, aka MLO or (HAM) I3NOO
//     Work:   University of Padova - Department of Physics
//             Via F. Marzolo, 8 - 35131 PADOVA - Italy
//     Phone:  +39 (049) 827-7216   FAX: +39 (049) 827-7102
//     EMail:  loreti@pd.infn.it
//     WWW:    http://www.pd.infn.it/~loreti/mlo.html
//
// -------------------------------------------------------------------
//
// (Non template) methods for conf.hh
//
// $Id: conf.cxx,v 1.4 2005/01/21 08:29:43 loreti Exp $
//
// -------------------------------------------------------------------

#include "conf.hh"

using std::cerr;
using std::cout;
using std::endl;
using std::isspace;
using std::ostream;
using std::string;

// The maximum length of a line to be read from setupMaster::readFile

const size_t setupMaster::lineLength = 128;

void setupMaster::readFile(
  std::ifstream     * inFile,
  const std::string & fileName
) {

  // readFile() reads all the lines of a configuration file; parses
  // the lines to find the keyword, then calls doConfig to perform the
  // assignments.
  
  char         buffer[lineLength];
  unsigned int lineNr = 0;

  while (inFile->getline(buffer, lineLength)) {
    ++lineNr;

    // For every line, strips the comments inserting a '\0' over the
    // first unquoted ';' found.

    char *pc, *pName, *pValue;

    for (pc = buffer,  pName = buffer + 1;   *pc;   ) {
      if (*pc == '\\'   &&   *pName == ';') {
        std::strcpy(pc, pName);
      } else if (*pc == ';') {
        *pc = '\0';
        break;
      }
      pc = pName++;
    }

    // Skips the leading whitespace; "pName" points to the first non
    // blank character (or to the trailing '\0' for an empty line).
    // The first non blank character cannot be a '='.

    for (pName = buffer;
         *pName && isspace(static_cast<unsigned char>(*pName));
         pName++) { }

    if (*pName == '\0') {
      continue;
    } else if (*pName == '=') {
      cerr << "*** Keyword missing in an assignment\n"
           << " -> File \"" << fileName << "\", line "
           << lineNr << endl;
      continue;
    }

    // Looks for the assignment operator '='; delimits the symbolic
    // name field.

    if ((pValue = std::strchr(pName, '=')) == 0) {
      for (pc = pName;
           *pc && !isspace(static_cast<unsigned char>(*pc));
           pc++) { }
      *pc = '\0';
      cerr << "*** No assignment to keyword \"" << pName << "\"\n"
           << " -> File \"" << fileName << "\", line "
           << lineNr << endl;
      continue;
    }
    *pValue++ = '\0';
    for (pc = pName;
         *pc && !isspace(static_cast<unsigned char>(*pc));
         pc++) { }
    *pc = '\0';

    // Skips the leading whitespace after the '=', to find the
    // beginning of the "value" field

    while (*pValue && isspace(static_cast<unsigned char>(*pValue))) ++pValue;
    if (*pValue == '\0') {
      cerr << "*** No value for keyword \"" << pName << "\"\n"
           << " -> File \"" << fileName << "\", line "
           << lineNr << endl;
      continue;
    }

    string keyword( pName );
    doConfig(keyword, pValue, fileName, lineNr);
  }

  if (!inFile->eof()) {
    cerr << "*** File \"" << fileName << "\": read error after line "
         << lineNr << endl;
  }
}

void setupMaster::doConfig(
  const string       & keyword,
  char               * pValue,
  const string       & fileName,
  const unsigned int   lineNr
) {

  // doConfig looks for the keyword in the setupMaster dictionary, and
  // tries to assign the wanted values through the internal method
  // stringToValue: that returns OK, or VERBOSE (an explicative
  // message has been printed), or SILENT (no message).

  bool                err = false;
  setupDict::iterator it;
  namedVariable       tmp(keyword);

  it = _dict.find( &tmp );

  if (it == _dict.end()) {
    cerr << "*** Keyword \"" << keyword << "\" is unknown\n";
    err = true;
  } else {
    namedVariable::eParse status = (*it)->stringToValue( pValue );

    if (status == namedVariable::SILENT) {
      cerr << "*** Invalid value \"" << pValue
           << "\" for keyword \"" << keyword << "\"\n";
    }
    if (status != namedVariable::OK) err = true;
  }

  if (err) {
    cerr << " -> File \"" << fileName
         << "\", line " << lineNr << endl;
  }
}

void setupMaster::out(
  ostream & o                   // = std::cout
) const {
  if (_dict.empty()) {
    o << "- No keywords defined" << endl;
  } else {
    setupDict::const_iterator it, itEnd = _dict.end();

    for (it = _dict.begin();   it != itEnd;   ++it) {
      register namedVariable * pNV = *it;
      cout << "- ";
      pNV->out(o);
      cout << endl;
    }
  }
}

ostream & operator << (
  ostream           & o,
  const setupMaster & s
) {
  s.out(o);
  return o;
}
