//     Author: Maurizio Loreti, aka MLO or (HAM) I3NOO
//     Work:   University of Padova - Department of Physics
//             Via F. Marzolo, 8 - 35131 PADOVA - Italy
//     Phone:  +39 (049) 827-7216   FAX: +39 (049) 827-7102
//     EMail:  loreti@pd.infn.it
//     WWW:    http://www.pd.infn.it/~loreti/mlo.html
//
// -------------------------------------------------------------------
//
// Another approach, more C++-ish, to the problem of variables to be
// configured at run time.  This approach is less cumbersome that the
// one used in config.h/config.cxx; and can use, with the generic
// programming (templates) approach, every user-defined abstract type
// (and not just the predefined types int, double, ...) .  The only
// conditions on the data type to be configured are the existence of
// the copy constructor; of the assignment operator; and of an
// extractor from an input stream.
//
// $Id: conf.hh,v 1.7 2006/06/20 08:48:49 loreti Exp $
//
// -------------------------------------------------------------------

#ifndef CONF_HH_
#define CONF_HH_

#include <algorithm>            // Standard includes
#include <cctype>
#include <cstring>
#include <fstream>
#include <functional>
#include <iostream>
#include <set>
#include <sstream>
#include <stdexcept>
#include <string>
#include <vector>

// Local functions: "getDelim"

namespace {
  char *getDelim(
    char **pStart,
    char  *sep
  ) {

    // The string pointed to by "pStart" is parsed by successive calls
    // to "getDelim" in fields, according to the separator characters
    // stored in the string "sep".  The string to be parsed is
    // modified, storing a \0 after the last non whitespace character
    // of every field; the returned value is a pointer to the first
    // non whitespace character of every field, or NULL at the end of
    // the string.  Empty fields (whitespace between separators) are
    // allowed.

    register char *retVal;
    register char *pc;
    register char *fb = 0;
    
    if (!pStart  ||  !(retVal = *pStart)  ||  !sep) return 0;
    
    while (*retVal  &&  std::isspace(static_cast<unsigned char>(*retVal)))
      retVal++;
    
    if (!*(pc = retVal)) return 0;
    
    while (*pc) {
      if (std::isspace(static_cast<unsigned char>(*pc))) {
        if (!fb) fb = pc;
      } else {
        if (std::strchr(sep, *pc)) {
          *pc++ = '\0';
          break;
        }
        fb = 0;
      }
      pc++;
    }
    
    if (fb) *fb = '\0';
    *pStart = pc;
    return retVal;
  }
}

// Forward declaration

class setupMaster;

// "namedVariable": base class for variables that may be configured at
// runtime.
//
// From "namedVariable" we later will derive the two template classes
// actually used to represent variables of an abstract data type T and
// to be configured at runtime: "setupVariable<T>" (scalar variables)
// and "setupArray<T>" (arrays).  Pointers to "namedVariable" will be
// used (in the class "setupMaster") as generic pointers to the
// instances of these template classes.
//
// The internal state of a "namedVariable" is described by a symbolic
// name (a string without spaces).  The methods include: comparison
// operators; an accessor that returns the name; and two other virtual
// methods for printouts and to change the value of the associated
// variable.
//
// Copy constructor and assignment operator are declared private and
// never defined in order to prevent their use; the virtual methods
// throw an exception (they cannot be used on the base class; but are
// not declared pure virtual, because we later will need to build a
// "namedVariable" from a std::string).
//
// An enumerated type ("eParse") is defined, that will be used for the
// return value of the virtual method "stringToValue": "OK" means "no
// errors; "VERBOSE" means "error, with an explicative message printed
// to stderr"; "SILENT" means "error, without messages".

class namedVariable {
protected:
  const std::string _name;

public:
  enum eParse { OK, SILENT, VERBOSE };

  namedVariable(const std::string & s) : _name(s) {}

  virtual ~namedVariable() {}

  bool operator < (const namedVariable & rhs) const {
    return _name < rhs._name; }

  bool operator == (const namedVariable & rhs) const {
    return _name == rhs._name; }

  const std::string & getName() const { return _name; }

  virtual void out(std::ostream & o = std::cout) const {
    throw std::runtime_error("Direct call to namedVariable::out"); }

  virtual eParse stringToValue(char * pc) {
    throw std::runtime_error("Direct call to namedVariable::stringToValue"); }

private:
  namedVariable(const namedVariable &);
  const namedVariable & operator = (const namedVariable &);

}; // class namedVariable

// "setupVariable": a template class derived from "namedVariable",
// that abstracts a scalar variable of type T.  T is an abstract type
// having working copy constructor, assignment operator and extractor
// from input streams.
//
// The associated methods are: constructor; a member function for
// printouts; and a procedure that reads a new value for the
// associated variable from a std::string.  That last method is
// private and may be called only from the friend class "setupMaster".
//
// Copy constructor and assignment operator are declared private and
// never defined in order to prevent their use.

template <typename T> class setupVariable : public namedVariable {
  friend class setupMaster;

private:
  T * _ptr;

public:
  setupVariable(const std::string & s, T * pT) : namedVariable(s), _ptr(pT) {}

  ~setupVariable() {}

  void out (std::ostream & o = std::cout) const {
    o << _name << " = " << *_ptr; }

private:
  setupVariable(const setupVariable &);
  const setupVariable & operator = (const setupVariable &);

  eParse stringToValue(char *);
}; // class setupVariable

// ostream output (operator <<) for setupVariable<T>

template <typename T> inline std::ostream & operator << (
  std::ostream           & o,
  const setupVariable<T> & s
) {
  s.out(o);
  return o;
}

// Conversion of a std::string to a variable of type T, associated to
// a setupVariable<T>

template <typename T> namedVariable::eParse setupVariable<T>::stringToValue(
  char * pValue
) {
  char * pc = getDelim(&pValue, ":");
  if (pc != 0   &&   *pc != '\0') {
    std::istringstream is( pc );
    char               c;
    T                  t;

    is >> t;
    if (is.fail()   ||   ((c = is.peek()) != EOF   &&   c != ' ')) {
      return SILENT;
    } else {
      *_ptr = t;
      return OK;
    }
  } else {
    std::cerr << "*** Assigning an empty field to keyword \""
              << getName() << "\" (ignored)\n";
    return VERBOSE;
  }
}

// The class "setupArray"; the equivalent of "setupVariable" for
// arrays.  "_num" is the array size.

template <typename T> class setupArray : public namedVariable {
  friend class setupMaster;

private:
  T            * _ptr;
  const size_t   _num;

public:
  setupArray(const std::string & s, T * pT, const size_t n)
    : namedVariable(s), _ptr(pT), _num(n) {}

  ~setupArray() {}

  void out (std::ostream & o = std::cout) const;

private:
  setupArray(const setupArray &);
  const setupArray & operator = (const setupArray &);

  eParse stringToValue(char *);
}; // class setupArray

// ostream output (member function) for a setupArray<T>

template <typename T> void setupArray<T>::out(
  std::ostream & o
  ) const {
  o << getName() << '[' << _num << "] = {";
  for (size_t i = 0;   i < _num;   ++i) {
    if (i) o << ", ";
    o << _ptr[i];
  }
  o << '}';
}

// ostream output (operator <<) for a setupArray<T>

template <typename T> inline std::ostream & operator << (
  std::ostream        & o,
  const setupArray<T> & s
) {
  s.out(o);
  return o;
}

// A function object used by setupArray<T>::stringToValue() for
// internal cleanup (calls delete on a pointer to a std::string)

struct delSP : public std::unary_function<std::string *, void> {
  void operator () (std::string * pS) { delete pS; }
};

// Conversion of an input string to an array of variables of type T,
// associated to a setupArray<T>

template <typename T> namedVariable::eParse setupArray<T>::stringToValue(
  char * pValue
) {
  std::vector< std::string * > * pStrings = new std::vector< std::string * >;
  eParse retVal = OK;

  char * pc;
  while ((pc = getDelim(&pValue, ":")) != 0) {
    std::string * pField = new std::string( pc );
    pStrings->push_back( pField );
  }

  size_t nMax = pStrings->size();
  if (nMax != _num) {
    std::cerr << "*** Warning: trying to configure " << getName()
              << '[' << _num << "] with " << nMax << " values\n";
    if (_num < nMax) nMax = _num;
    retVal = VERBOSE;
  }

  std::istringstream is;
  std::vector< std::string * >::const_iterator it = pStrings->begin();
  for (size_t i = 0;   i < nMax;   ++i,  ++it) {
    is.str( *(*it) );

    char c;
    T    t;
    is >> t;

    if (is.fail()   ||   ((c = is.peek()) != EOF   &&   c != ' ')) {
      std::cerr << "*** Invalid value \"" << is.str()
                << "\" for keyword \"" << getName()
                << '[' << i << "]\n";
      retVal = VERBOSE;
    } else {
      _ptr[i] = t;
    }
    is.clear();
  }

  std::for_each(pStrings->begin(), pStrings->end(), delSP());
  delete pStrings;
  return retVal;
}

// A function object that defines a strict weak ordering; will be used
// to store pointers to "namedVariables" according to the ascending
// order of the associated keyword.

struct comparePointedNamedVariables
  : public std::binary_function<const namedVariable *,
                                const namedVariable *,
                                bool> {
  bool operator () (const namedVariable *p1, const namedVariable *p2) const {
    return *p1 < *p2;
  }
};
 
// "setupDict" is defined as a shortcut for a set of pointers to
// namedVariables, ordered by the above function object.
     
typedef std::set<namedVariable *, comparePointedNamedVariables> setupDict;

// A function object used by the destructor of the class setupMaster
// for internal cleanup (calls operator delete for a pointer to a
// namedVariable).

struct delNVP : public std::unary_function<namedVariable *, void> {
  void operator () (namedVariable * pNV) { delete pNV; }
};

// A "setupMaster" owns a "setupDict"; and has methods to add a
// namedVariable and to read a configuration file.  Other methods are
// used internally.

class setupMaster {
public:
  static const size_t lineLength;

private:
  setupDict _dict;

public:
  setupMaster() {}
  ~setupMaster() { std::for_each(_dict.begin(), _dict.end(), delNVP()); }
  template <typename T> void add(const std::string &, T *, const size_t = 1);
  void readFile(std::ifstream *, const std::string &);
  void out(std::ostream & o = std::cout) const;

private:
  void doConfig(const std::string &, char *,
                const std::string &, const unsigned int);
}; // class setupMaster

std::ostream & operator << (std::ostream &, const setupMaster &);

// Adds a namedVariable to the internal dictionary

template <typename T> void setupMaster::add(
  const std::string & keyword,
  T                 * pT,
  const size_t        n         // = 1
) {
  namedVariable * pNV;
  if (n == 1) {
    pNV = new setupVariable<T>(keyword, pT);
  } else {
    pNV = new setupArray<T>(keyword, pT, n);
  }

  std::pair< setupDict::iterator, bool > pib = _dict.insert( pNV );

  if (!pib.second) {
    std::cerr << "*** The keyword \"" << keyword
              << "\" is already known (the new definition is ignored)\n";
    delete pNV;
  }
}

// Two preprocessor macros to ease the use of the above class:
// - CONFVAR adds to the SetupMaster instance "sm" the scalar variable
//   "x" with name "x";
// - CONFARRAY adds to the SetupMaster instance "sm" the array
//   variable "x" with name "x".

#define CONFVAR(sm, x)   sm.add(#x, &x)
#define CONFARRAY(sm, x) sm.add(#x, x, sizeof(x)/sizeof(x[0]))

#endif // CONF_HH_
