/*------------------------------------------------------*
 | Author: Maurizio Loreti, aka MLO or (HAM) I3NOO      |
 | Work:   University of Padova - Department of Physics |
 |         Via F. Marzolo, 8 - 35131 PADOVA - Italy     |
 | Phone:  +39 (049) 827-7216   FAX: +39 (049) 827-7102 |
 | EMail:  loreti@padova.infn.it                        |
 | WWW:    http://wwwcdf.pd.infn.it/~loreti/mlo.html    |
 *------------------------------------------------------*

 Routines to build and search a table of keywords looking
 for minimal length abbreviations.

 $Id: keyword.c,v 1.4 2002/12/17 06:42:19 loreti Exp $

 *------------------------------------------------------*/

#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "keyword.h"

#ifdef TEST
#include <stdio.h>
int main()                      /* Quick and dirty. Sorry. */
{
  char *zz[] = {
    "EOJ", "PATCH", "COPY", "ANALYZE",
    "PURGE", "RECOVER", "RECALL", "EXAMINE",
    "STOP", "DEBUG", "JOIN", "RUN"
  };
  KeyTable *pKT;
  KeyErr KE;
  int i, n = sizeof(zz)/sizeof(char *);

  pKT = initKeyTable(n, zz, 0, &KE);
  if (pKT) {
    for (i=0; i<n; i++) {
      printf("Key %d (n = %d): %s\n", i, pKT->start[i].n, pKT->start[i].id);
    }
    for (;;) {
       char string[128];
       int pi;

       printf("String: ");
       if (fgets(string, 128, stdin)) {
         char *pc;

         if ((pc = strchr(string, '\n')) != 0) *pc = '\0';
         pi = getKeyword(pKT, string, &KE);
         if (KE == KE_OK) {
           printf("==> %s\n", zz[pi]);
         } else {
           switch (KE) {
             case KE_UNDEF:   puts("Undefined keyword"); break;
             case KE_AMBIG:   puts("Ambiguous keyword"); break;
             case KE_INVDATA: puts("Invalid input");     break;
             default:         puts("Can't happen");      break;
           }
         }
       } else {
         break;
       }
    }
    releaseKeyTable(pKT);
  } else {
    switch (KE) {
      case KE_INVDATA: puts("Invalid table");      break;
      case KE_NOMEM:   puts("No memory vailable"); break;
      case KE_DUPL:    puts("Duplicate key");      break;
      case KE_AMBIG:   puts("Ambiguous key");      break;
      default:         puts("Can't happen");       break;
    }
  }
  exit(EXIT_SUCCESS);
}
#endif /* TEST */

/**
 | Local procedures; strnicmp is in the Amiga SAS-C library.
**/

static void QuickSort(Key K[], int left, int right);
static void Swap(Key *K, int i, int j);
static void strlcpy(char *to, const char *from);

#ifndef __SASC
static int strnicmp(const char *s1, const char *s2, size_t n);
#endif

/*--------------------------------------------------------------------*/

KeyTable *initKeyTable(
  size_t  n,                    /* Number of keywords          */
  char   *key[],                /* Array of keywords           */
  int     user[],               /* Array of user data, or NULL */
  KeyErr *error                 /* Return error code           */
){

/*--------------------------------------------------------------*
 | Initialisation of a keyword table.  A keyword table is       |
 | organized as an (alphabetically ordered) array of 'Key'      |
 | structures: each of them contains the keyword string         |
 | (converted to lowercase), the minumum number of characters   |
 | needed to univocally identify the keyword in the context     |
 | of this table, and an user supplied integer (if 'user' is    |
 | NULL, the sequential number of the string in the array       |
 | 'key' is used).                                              |
 | The returned value is a pointer to the 'KeyTable' structure, |
 | or (if an error is detected), NULL.  'error' contains:       |
 | KE_OK:       all OK:                                         |
 | KE_INVDATA:  if n is zero;                                   |
 | KE_NOMEM:    for errors from malloc();                       |
 | KE_DUPL:     if the 'key' array contains twice the same key; |
 | KE_AMBIG:    if two of the keys are intrinsecally ambiguous, |
 |              e.g. 'foo' and 'foobar' together.               |
 *--------------------------------------------------------------*/

  KeyTable *pKT = 0;            /* Return value */

  if (n < 1) {                  /* Note: n is an unsigned type (size_t) */
    *error = KE_INVDATA;
  } else {
    if ((pKT = malloc(sizeof(KeyTable))) != 0) {
      if ((pKT->start = calloc(n, sizeof(Key))) != 0) {
        size_t i;
        Key *pK, *last;
        char **ppc;

/**
 | Build the table in the same order as 'key', then
 | sorts its elements.
**/

        pKT->n = n;
        for (i=0, pK=pKT->start, ppc=key;  i<n;  i++, pK++, ppc++) {
          size_t length = strlen(*ppc) + 1;

          if ((pK->id = malloc(length)) == 0) {
            *error = KE_NOMEM;
            goto mess;
          }
          strlcpy(pK->id, *ppc);
          pK->userdata = user ? user[i] : i;
        }
        QuickSort(pKT->start, 0, n-1);

/**
 | Scans the table comparing each entry with the next ones, to
 | compute the minimum number of characters required for a match.
**/

        for (i=0, last=0, pK=pKT->start;  i<n;  i++, last=pK, pK++) {
          if (last) {
            char *p1 = last->id;
            char *p2 = pK->id;

            while (*p1 == *p2) {
              if (*p1 == '\0') {
                *error = KE_DUPL;
mess:           releaseKeyTable(pKT);
                return 0;
              }
              p1++;
              p2++;
            }

            if (*p1 == '\0') {
              *error = KE_AMBIG;
              goto mess;
            }

            pK->n = ++p2 - pK->id;
            if (pK->n > last->n) {
              last->n = pK->n;
            }
          } else {
            pK->n = 1;
          }
        }
        *error = KE_OK;
      } else {
        free(pKT);
        *error = KE_NOMEM;
        return 0;
      }
    } else {
      *error = KE_NOMEM;
    }
  }
  return pKT;
}

int getKeyword(
  KeyTable *table,              /* Keyword table         */
  char     *string,             /* String to be compared */
  KeyErr   *error               /* Error flag            */
){

/*------------------------------------------------------------*
 | getKeyword compares 'string' with the keywords of 'table', |
 | and returns the user-supplied value associated with the    |
 | selected keyword.  On exit, 'error' contains:              |
 | KE_OK:       no error, all OK;                             |
 | KE_INVDATA:  NULL pointer for 'table', or invalid table;   |
 | KE_UNDEF:    no match;                                     |
 | KE_AMBIG:    match with more than a single keyword.        |
 *------------------------------------------------------------*/

  if (table   &&   table->start) {
    int    first  = 0;
    int    last   = table->n - 1;
    size_t length = strlen(string);

/**
 | Binary search in the keyword array for a match with 'string'
**/

    while (first <= last) {
      size_t n = (first + last) / 2;
      Key *pK  = table->start + n;
      int i;

      if ((i = strnicmp(string, pK->id, length)) == 0) {
        if (length < pK->n) {
          *error = KE_AMBIG;
          return 0;
        } else {
          *error = KE_OK;
          return pK->userdata;
        }
      } else if (i < 0) {
        last = n - 1;
      } else {
        first = n + 1;
      }
    }
    *error = KE_UNDEF;
  } else {
    *error = KE_INVDATA;
  }
  return 0;
}

void releaseKeyTable(
  KeyTable *table               /* Keyword table */
){

/**
 | Returns to the heap all the allocated storage.
**/

  if (table != 0) {
    if (table->start != 0) {
      int i;
      Key *pK;

      for (i=0, pK=table->start;  i<table->n;  i++, pK++) {
        if (pK->id != 0) free(pK->id);
      }
      free(table->start);
    }
    free(table);
  }
}

static void QuickSort(
  Key K[],                      /* Array to be sorted, */
  int left,                     /*   from K[left]      */
  int right                     /*     to K[right].    */
){

/**
 | Copied, with Swap(), from Kernighan and Ritchie - "The C
 | programming language", 2nd Ed. (Addison-Wesley) - pag. 110.
**/

  int i, last;

  if (left < right) {
    Swap(K, left, (left + right)/2);
    last = left;
    for (i=left+1; i<=right; i++) {
      if (strcmp(K[i].id, K[left].id) < 0) {
        Swap(K, ++last, i);
      }
    }
    Swap(K, left, last);
    QuickSort(K, left, last-1);
    QuickSort(K, last+1, right);
  }
}

static void Swap(
  Key *K,
  int  i,
  int  j
){
  if (i != j) {
    Key temp;

    (void) memcpy(&temp, K+i,  sizeof(Key));
    (void) memcpy(K+i,   K+j,  sizeof(Key));
    (void) memcpy(K+j,  &temp, sizeof(Key));
  }
}

static void strlcpy(
  char       *to,               /* Destination string */
  const char *from              /* Source string      */
){

/**
 | Copies 'from' to 'to', converting to lowercase.  The loop is NOT
 | written with "tolower(*from++)" against possible side effects in
 | some tolower macro implementations.
**/

  while ((*to++ = tolower((unsigned char) *from)) != '\0') {
    from++;
  }
}

#ifndef __SASC
static int strnicmp(
  const char *s1,               /* Compare n characters  */
  const char *s2,               /*   of two strings,     */
  size_t n                      /*     case insensitive. */
){
  while (n--) {
    int i;

    i = tolower((unsigned char) *s1) - *s2;
    if (i  ||  *s2 == '\0') return i;
    s1++;                               /* See the above comment... */
    s2++;
  }
  return 0;
}
#endif /* __SASC */
